
#ifndef EL__ECMASCRIPT_SPIDERMONKEY_WINDOW_H
#define EL__ECMASCRIPT_SPIDERMONKEY_WINDOW_H

#include "ecmascript/spidermonkey/util.h"

extern const JSClass window_class;
extern const JSPropertySpec window_props[];
extern const spidermonkeyFunctionSpec window_funcs[];

#endif
