
#ifndef EL__ECMASCRIPT_SPIDERMONKEY_NAVIGATOR_H
#define EL__ECMASCRIPT_SPIDERMONKEY_NAVIGATOR_H

#include "ecmascript/spidermonkey/util.h"

extern const JSClass navigator_class;
extern const JSPropertySpec navigator_props[];

#endif
